/* STRETCHY - sample code for a very basic time-domain expansion instrument

   p0 = output start time
   p1 = input start time
   p2 = input duration
   p3 = amplitude multiplier
   p4 = number of times to repeat each buffer
   p5 = envelope table
   p6 = input channel [optional, default is 0]
   p7 = pan (in percent-to-left format) [optional, default is 0]
*/

#include <stdio.h>
#include <stdlib.h>
#include <ugens.h>
#include <Ougens.h>
#include "STRETCHY.h"          // declarations for this instrument class
#include <rt.h>
#include <rtdefs.h>

STRETCHY::STRETCHY()
    : inbuffer(NULL), resetter(0)
{
}

STRETCHY::~STRETCHY()
{
    delete [] inbuffer;
}

int STRETCHY::init(double p[], int n_args)
{
    if (rtsetoutput(p[0], p[2], this) == -1)
		return DONT_SCHEDULE;

	if (outputChannels() > 2)
		return die("STRETCHY", "Use mono or stereo output only.");

	if (rtsetinput(p[1], this) == -1)
		return DONT_SCHEDULE;

    inchan = p[6];
	if (inchan >= inputChannels())
		return die("STRETCHY", "You asked for channel %d of a %d-channel input.",
		                                             inchan, inputChannels());

    // set up our input object
    theInput = new Ortgetin(this);
    
    // get the envelope table stuff for each sample chunk
    thenvelope = (double *)getPFieldTable(5, &thenvelopelength); 
    
    amp = p[3];
    stretchfactor = p[4];
    stretchcounter = 0;
    pan = p[7];

	return nSamps();
}


// Allocate the input buffer.  For non-interactive (script-driven) sessions,
// the constructor and init() for every instrument in the script are called
// before any of them runs.  By contrast, configure() is called right before
// the instrument begins playing.  If we were to allocate memory at init
// time, then all notes in the score would allocate memory then, resulting
// in a potentially excessive memory footprint.
int STRETCHY::configure()
{
    // RTBUFSAMPS is the maximum number of sample frames processed for each
	// call to run() below.    
	inbuffer = new float [RTBUFSAMPS];
    
	return inbuffer ? 0 : -1;	// IMPORTANT: Return 0 on success, and -1 on failure.
}


// Called at the control rate to update parameters like amplitude, pan, etc.

void STRETCHY::doupdate()
{
	// The Instrument base class update() function fills the <p> array with
	// the current values of all pfields.  There is a way to limit the values
	// updated to certain pfields.  For more about this, read
	// src/rtcmix/Instrument.h.

	double p[8];
	update(p, 8);

	amp = p[3];
    pan = p[7];
}


// Called by the scheduler for every time slice in which this instrument
// should run.  This is where the real work of the instrument is done.

int STRETCHY::run()
{
    float out[2];
    float in[2];

    // load up the buffer if we need a new load
    if (--stretchcounter <= 0) {
        float thenvelopelocation = 0.0;
        float thenvelopeskipper = (float)thenvelopelength/(float)framesToRun();
        
        for (int i = 0; i < framesToRun(); i++) {
            theInput->next(in);
            inbuffer[i] = in[inchan] * thenvelope[(int)thenvelopelocation];
            thenvelopelocation += thenvelopeskipper;
        }
        stretchcounter = stretchfactor;
    }
    
	// now write out that buffer
	for (int i = 0; i < framesToRun(); i++) {
		if (--resetter <= 0) {
			doupdate();
			resetter = getSkip();
		}

		// Grab the current input sample, scaled by the amplitude multiplier.
		out[0] = inbuffer[i] * amp;

		// If we have stereo output, use the pan pfield.
		if (outputChannels() == 2) {
			out[1] = out[0] * (1.0 - pan);
			out[0] *= pan;
		}

		// Write this sample frame to the output buffer.
		rtaddout(out);

		// Increment the count of sample frames this instrument has written.
		increment();
	}

	// Return the number of frames we processed.
	return framesToRun();
}


// The scheduler calls this to create an instance of this instrument
// and to set up the bus-routing fields in the base Instrument class.
// This happens for every "note" in a score.

Instrument *makeSTRETCHY()
{
	STRETCHY *inst = new STRETCHY();
	inst->set_bus_config("STRETCHY");

	return inst;
}


// The rtprofile introduces this instrument to the RTcmix core, and
// associates a script function name (in quotes below) with the instrument.
// This is the name the instrument goes by in a script.

void rtprofile()
{
	RT_INTRO("STRETCHY", makeSTRETCHY);
}


